package com.blogspot.revbingo.threed;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.geom.AffineTransform;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import javax.swing.JPanel;

import com.blogspot.revbingo.threed.lights.LightScene;
import com.blogspot.revbingo.threed.objects.BasicSceneObject;

/*
 * A panel that will display our 3D scene
 */
public class ThreeDeePanel extends JPanel {

	ArrayList<BasicSceneObject> worldScene = null;  //holds the list of all objects to be drawn
	LightScene lights = null;
	private AffineTransform modelToDeviceTransform = null;
	
	private InfoPanel infoPanel;
	
	public InfoPanel getInfoPanel() {
		return infoPanel;
	}

	public void setInfoPanel(InfoPanel info) {
		this.infoPanel = info;
	}

	public void setScene(ArrayList<BasicSceneObject> s, LightScene lights){
		worldScene = s;
		this.lights = lights;
	}
	
	/*
	 * Our transform uses getWidth() and getHeight(), which can change when the panel is resized. To
	 * deal with this, we override setBounds and recreate the AffineTransform.
	 *  The constructor for AffineTransform essentially defines
	 * a 3x3 matrix  see the javadoc for that class for more info
	 * 
	 * | 1  0  width/2 |
	 * | 0 -1 height/2 |
	 * | 0  0    1     |
	 * 
	 * (the last row is implicit)
	 *  
	 * The result is that the Y axis is inverted (up is +ve), and the origin is in the centre of 
 	 * the screen
	 */
	@Override
	public void setBounds(int x, int y, int width, int height) {
		modelToDeviceTransform = new AffineTransform(1,0,0,-1,this.getWidth()/2,this.getHeight()/2);		
		super.setBounds(x, y, width, height);
	}

	@Override
	public void setBounds(Rectangle r) {
		modelToDeviceTransform = new AffineTransform(1,0,0,-1,this.getWidth()/2,this.getHeight()/2);
		super.setBounds(r);
	}

	@Override
	protected void paintComponent(Graphics g) {
		super.paintComponent(g);	//standard stuff
	
		if(worldScene == null) return;  //don't try and draw stuff if it's not ready yet
		
		long start = System.nanoTime();
		Graphics2D g2 = (Graphics2D) g;
		g2.setColor(Color.WHITE);	//we'll draw everything in white to start
		
		//this ArrayList holds the polygons that are actually to be rendered
		List<Primitive> renderScene = new LinkedList<Primitive>();
		
		//combine polygons from all objects into scene.  We need to mix them all together
		//so that they can be sorted for z order
		for(BasicSceneObject object : worldScene){
			renderScene.addAll(object.getPolygons());
		}
		
		//do all the main pipeline actions
		doPipeline(renderScene);
		
		//convert scene coordinates into coords suitable for the display
		//this transform is applied on each call to draw()
		g2.transform(modelToDeviceTransform);
		
		//project the coordinates to the screen and draw
		for(Primitive poly : renderScene){
			poly.project();
			poly.draw(g2);
		}
		
		infoPanel.draw(g2);
		
		long end = System.nanoTime();
		long renderTime = (end - start);
		infoPanel.setPolyCount(renderScene.size());
		infoPanel.setFps(1000000000d/renderTime);
	}

	private void doPipeline(List<Primitive> scene) {
		//cull where we can
		backfaceCulling(scene);
		
		//sort the scene by z-order, see SceneObject.compareTo()
		sortForZOrder(scene);
	
		//light the polygons
		lightScene(scene);
	}

	private void backfaceCulling(List<Primitive> scene) {
		Iterator<Primitive> i = scene.iterator();
		//Iterate through the scene and remove and polygons that don't
		//need to be drawn
		while(i.hasNext()){
			if(i.next().canBeCulled()) i.remove();
		}
	}
	
	private void lightScene(List<Primitive> scene) {
		for(Primitive p : scene) {
			p.lightPolygon(lights);
		}
	}
	
	private void sortForZOrder(List<Primitive> scene) {
		Collections.sort(scene);
	}
	
}
