package com.blogspot.revbingo.threed;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.color.ColorSpace;

import com.blogspot.revbingo.threed.lights.Light;
import com.blogspot.revbingo.threed.lights.LightScene;

public class Triangle extends Primitive {

	public static boolean WIREFRAME = false;
	
	//vertices of the triangle in world space
	double[] x = new double[3];
	double[] y = new double[3];
	double[] z = new double[3];
	
	//points projected on to the screen
	private int[] xPoints = new int[3];
	private int[] yPoints = new int[3];

	//direction of the camera
	private Vector viewer = new Vector(0,0,1);
	
	//the vector normal of this triangle.  Previously this was being
	//calculated every frame, but it only needs recalculating if the
	//polygon is rotated.
	private Vector normal;
	
	//color of the polygon when lit
	private Color litColor;
	//color of the surface
	private Color surfaceColor = Color.WHITE;
	

	public Triangle(double p1x, double p1y, double p1z, double p2x, double p2y,
			double p2z, double p3x, double p3y, double p3z) {
		x[0] = p1x;
		x[1] = p2x;
		x[2] = p3x;

		y[0] = p1y;
		y[1] = p2y;
		y[2] = p3y;

		z[0] = p1z;
		z[1] = p2z;
		z[2] = p3z;
		
		normal = getNormal().normalise();
	}
	
	public void setSurfaceColor(Color c) {
		surfaceColor = c;
	}

	/*
	 * Determines if this polygon can be culled from the scene, based on the direction
	 * of it's vector normal.
	 * 
	 * If we're showing the scene in wireframe mode, we don't cull
	 */
	public boolean canBeCulled() {
		if(WIREFRAME) return false;
		
		double cull = normal.dotProduct(viewer);
		return (cull > 0);
	}
	
	/*
	 * Applies lighting to the polygon.
	 */
	@Override
	public void lightPolygon(LightScene lights) {
		/*
		 * Loop through all the lights in the scene and work out how
		 * they affect this polygon
		 */
		litColor = Color.BLACK;
		for(Light light : lights) {
				litColor = addColor(litColor, light.light(this));
		}
		litColor = multiplyColor(litColor, surfaceColor);
	}
	
	private Color addColor(Color c1, Color c2) {
		return new Color(Math.min(255, c1.getRed() + c2.getRed()),
				Math.min(255, c1.getGreen() + c2.getGreen()), 
				Math.min(255, c1.getBlue() + c2.getBlue()));
	}
	
	private Color multiplyColor(Color c1, Color c2) {
		float[] c1Comp = c1.getColorComponents(null);
		float[] c2Comp = c2.getColorComponents(null);
		return new Color(c1Comp[0] * c2Comp[0], 
						c1Comp[1] * c2Comp[1],
						c1Comp[2] * c2Comp[2]);
	}
	
	/*
	 * Project world space coordinates to screen coords
	 */
	public void project() {
		for (int i = 0; i < 3; i++) {
			xPoints[i] = (int) x[i];
			yPoints[i] = (int) y[i];
		}
	}
	
	/*
	 * Draw the polygon.  If in wireframe mode, simply render outline in white,
	 * otherwise draw flat-shaded polygon 
	 */
	public void draw(Graphics2D graphics) {
		
		if(WIREFRAME) {
			graphics.drawPolygon(xPoints, yPoints, 3);
		} else {
			graphics.setColor(litColor);
			graphics.fillPolygon(xPoints, yPoints, 3);
		}
	}
	
	/*
	 * Translate this polygon to another point in world space
	 */
	public void translate(double dx, double dy, double dz) {
		for (int i = 0; i < 3; i++) {
			x[i] += dx;
			y[i] += dy;
			z[i] += dz;
		}
		//note that the direction of the polygon has not changed, so no need to 
		//recalculate the normal
	}
	
	/*
	 * Rotate this polygon in world space by the given angle, and around the specified axis
	 */
	public void rotate(double radians, RotationAxis axis) {
		double[] a = null;
		double[] b = null;
		
		//This switch determines which points we rotate in the equation
		//at the bottom
		switch(axis) {
			case X:
				a = y;
				b = z;
				break;
			case Y:
				a = z;
				b = x;
				break;
			case Z:
				a = x;
				b = y;
				break;
		}
		
		double[] sa = a.clone();
		for (int i = 0; i < 3; i++) {
			a[i] = (sa[i] * Math.cos(radians)) - (b[i] * Math.sin(radians));
			b[i] = (b[i] * Math.cos(radians)) + (sa[i] * Math.sin(radians));
		}
		
		//Position changed, so recalculate the vector normal
		normal = getNormal().normalise();
	}
	
	public void rotate(RotationMatrix rotationMatrix) {
		Matrix[] points = new Matrix[3];
		
		for(int i=0;i<3;i++) {
			points[i] = new Matrix(new double[][] {{x[i]}, {y[i]}, {z[i]}});
			Matrix result = rotationMatrix.times(points[i]);
			x[i] = result.get(0,0);
			y[i] = result.get(1,0);
			z[i] = result.get(2,0);
		}
		//Position changed, so recalculate the vector normal
		normal = getNormal().normalise();
	}

	/*
	 * Return the vector normal of this polygon
	 */
	public Vector getNormal() {

		Point p1 = new Point(x[0], y[0], z[0]);
		Point p2 = new Point(x[1], y[1], z[1]);
		Point p3 = new Point(x[2], y[2], z[2]);

		Vector edge1 = p2.vectorTo(p1);
		Vector edge2 = p3.vectorTo(p2);

		return edge2.crossProduct(edge1);
	}

	/*
	 * Return the z-order of this polygon - an average of the z-distance
	 * for each vertex
	 */
	@Override
	public double getZOrder() {
		return (z[0] + z[1] + z[2]) / 3;
	}

	@Override
	public Point getPosition() {
		return new Point((x[0] + x[1] + x[2])/3, (y[0] + y[1] + y[2])/3, (z[0] + z[1] + z[2])/3);
	}
	
	
	
}
